import pygame
from pygame import Rect
from resource import get_image, get_sound
from utils import make_array
from level_io import Saveable, register_class
from animation import Animation

cell_size = (60, 120)
cell_depth = (20, 20)
wall_thickness = 10
floor_thickness = 10

cell_width, cell_height = cell_size
cell_hdepth, cell_vdepth = cell_depth

#cell_rect = Rect((0, 0), cell_size)
#wall_rect = Rect((0, 0),
#	(cell_depth[0] + wall_thickness, cell_depth[1] + cell_size[1]))
#floor_rect = Rect((0, 0),
#	(cell_depth[0] + cell_size[0], cell_depth[1] + floor_thickness))

class Level(object):
	#  num_rows   int
	#  num_cols   int
	#  backs      array[nrows, ncols] of Back
	#  gizmos     array[nrows, ncols] of Gizmo
	#  objects    array[nrows, ncols] of Object
	#  walls      array[nrows, ncols+1] of Wall
	#  floors     array[nrows+1, ncols] of Floor
	#  actors     list of Actor, including player
	#  player     Player
	
	def __init__(self, nrows, ncols):
		self.num_rows = nrows
		self.num_cols = ncols
		self.cells = make_array(None, nrows, ncols, 3)
		self.backs = self.cells[:,:,0]
		self.gizmos = self.cells[:,:,1]
		self.objects = self.cells[:,:,2]
		self.walls = make_array(None, nrows, ncols + 1)
		self.floors = make_array(None, nrows + 1, ncols)
		self.actors = []
		self.player = None
		self.filename = None
	
	def set_player(self, player):
		if self.player:
			self.actors.remove(self.player)
		self.player = player
		self.actors.append(player)

	def draw(self, surface):
		cw, ch = cell_size
		dw, dh = cell_depth
		nrows = self.num_rows
		ncols = self.num_cols
		cell_rows = xrange(self.num_rows - 1, -1, -1)
		cell_cols = xrange(self.num_cols)
		#floor_rows = xrange(self.num_rows, -1, -1)
		#wall_cols = xrange(self.num_cols + 1)
		backs = self.backs
		cells = self.cells
		walls = self.walls
		floors = self.floors
		gizmos = self.gizmos
		
		# Walls with doors in them are split into a front and back image.
		# If behind is true, render the back half, otherwise render the
		# front (or only) half.
	
		def render_back(row, col):
			back = backs[row, col]
			if back:
				r = Rect(0, 0, dw, ch + dh)
				r.topright = dw + (col + 1) * cw, row * ch
				back.draw(surface, r)

		def render_wall(row, col, behind):
			wall = walls[row, col]
			if wall:
				r = Rect(0, 0, dw, ch + dh)
				r.topright = dw + col * cw, row * ch
				if behind:
					wall.draw_behind(surface, r)
				else:
					wall.draw(surface, r)
		
		def render_floor(row, col, front):
			floor = floors[row, col]
			if floor:
				r = Rect(0, 0, cw + dw, dh)
				r.topright = dw + (col + 1) * cw, row * ch
				floor.draw(surface, r, front)
		
		def render_cell(row, col):
			back, gizmo, obj = cells[row, col]
			r = Rect(col * cw, row * ch, cw + dw, ch + dh)
			if gizmo:
				gizmo.draw(surface, r, 0)
			if obj:
				obj.draw(surface, r)
		
		def render_gizmo_front(row, col):
			gizmo = gizmos[row, col]
			if gizmo and gizmo.front_image:
				r = Rect(col * cw, row * ch, cw + dw, ch + dh)
				gizmo.draw(surface, r, 1)

		player = self.player
		if player:
			# Set prow, pcol to the cell in which the player should be rendered.
			# This is not necessarily the same as the one nominally containing the
			# player. When climbing from a cell to the one above, the player needs to
			# be drawn after the background of both cells. But if there is a solid
			# floor above the player, the floor needs to be drawn after the player
			# and before anything in the cell above.
			prow, pcol = self.player.cell(self)
			if prow > 0:
				floor = self.floor_at(prow, pcol)
				if not floor or floor.passable():
					prow -= 1
			pz = player.z
		else:
			prow = pcol = None
			pz = 0
		#print "Begin draw level"
		#print "Drawing player in cell", prow, pcol
		# We render in columns, from bottom to top and left to right.
		#col = 0
		# Render all of the cell backgrounds
		for col in cell_cols:
			for row in cell_rows:
				render_back(row, col)
		# Render the leftmost column of walls
		for row in cell_rows:
			render_wall(row, 0, 0)
		# Render remaining columns of cells and walls to their right
		for col in cell_cols:
			render_floor(nrows, col, 0)
			for row in cell_rows:
				render_cell(row, col)
				render_wall(row, col+1, 1)
				if pz >= 0:
					render_gizmo_front(row, col)
				if row == prow and col == pcol:
					self.player.draw(surface, self)
				if row+1 < nrows:
					if pz < 0:
						render_gizmo_front(row+1, col)
					if not floors[row+1, col]:
						render_wall(row+1, col+1, 0)
				render_floor(row+1, col, 1)
				if row == 0 or floors[row, col]:
					render_wall(row, col+1, 0)
					render_floor(row, col, 0)
			if pz < 0:
				render_gizmo_front(0, col)
			render_floor(0, col, 0)
			render_floor(0, col, 1)
		#print "End draw level"

	def actor_pos_to_cell_coords(self, (x, y)):
		cw, ch = cell_size
		col, dx = divmod(x, cw)
		row, dy = divmod(y, ch)
		return (self.num_rows - row - 1, col), (dx, dy)
	
	def cell_coords_to_actor_pos(self, (row, col), (dx, dy)):
		cw, ch = cell_size
		x = col * cw + dx
		y = (self.num_rows - row - 1) * ch + dy
		return (x, y)

	def begin_frame(self):
		nrows = self.num_rows
		ncols = self.num_cols
		#backs = self.backs
		cells = self.cells
		walls = self.walls
		floors = self.floors
		for row in xrange(nrows + 1):
			for col in xrange(ncols + 1):
				coords = row, col
				if row < nrows and col < ncols:
					for item in cells[coords]:
						if item:
							item.begin_frame(self, coords)
					#back = backs[coords]
					#if back:
					#	back.begin_frame(self, coords)
				if row < nrows:
					wall = walls[coords]
					if wall:
						wall.begin_frame(self, coords)
				if col < ncols:
					floor = floors[coords]
					if floor:
						floor.begin_frame(self, coords)
		for actor in self.actors:
			actor.begin_frame(self)
	
	def wall_at(self, row, col):
		if 0 <= row < self.num_rows and 0 <= col <= self.num_cols:
			return self.walls[row, col]
		else:
			return None
		
	def floor_at(self, row, col):
		if 0 <= row <= self.num_rows and 0 <= col < self.num_cols:
			return self.floors[row, col]
		else:
			return None
	
	def gizmo_at(self, row, col):
		if 0 <= row < self.num_rows and 0 <= col < self.num_cols:
			return self.gizmos[row, col]
		else:
			return None
	
	def object_at(self, row, col):
		if 0 <= row < self.num_rows and 0 <= col < self.num_cols:
			return self.objects[row, col]
		else:
			return None
	
	def take_object_at(self, row, col):
		obj = self.object_at(row, col)
		if obj:
			self.objects[row, col] = None
		return obj
	
	def locate_player(self, coords):
		pos = self.cell_coords_to_actor_pos(coords, (30, 0))
		self.player.position = pos
	
	def update_tags(self):
		tags = {}
		for row in xrange(self.num_rows):
			for col in xrange(self.num_cols):
				item = self.gizmos[row, col]
				if item:
					tag = getattr(item, 'tag', None)
					if tag:
						tags[tag] = ((row, col), item)
		self.tags = tags
	
	def find_tag(self, tag):
		try:
			return self.tags[tag]
		except KeyError:
			return (0, 0), None
		

class Item(Saveable):
	#  image_name   string
	#  image        surface
	
	image_name = None
	image = None
	border = 0
	front_image_name = None
	front_image = None
	front_border = 0
	is_passable = False
	thumbnail = None
	
	def __init__(self, image = None, **kwds):
		Saveable.__init__(self, image = image, **kwds)
	
	def _get_params(self, p):
		p.image = self.image_name
		if self.border:
			p.border = self.border
		if self.is_passable <> self.__class__.is_passable:
			p.passable = self.is_passable
		if self.front_image_name:
			p.front_image = self.front_image_name
		if self.front_border:
			p.front_border = self.front_border
	
	def _set_params(self, p):
		self.image_name = getattr(p, 'image', None)
		self.border = getattr(p, 'border', 0)
		if self.image_name:
			self.image = get_image(self.image_name, border = self.border)
		self.front_image_name = getattr(p, 'front_image', None)
		self.front_border = getattr(p, 'front_border', None)
		if self.front_image_name:
			self.front_image = get_image(self.front_image_name, border = self.front_border)
		self.is_passable = getattr(p, 'passable', self.__class__.is_passable)
		#self.thumbnail = pygame.transform.rotozoom(self.image, 0, 0.5)

	def draw(self, surface, rect, front = False):
		if front:
			image = self.front_image
		else:
			image = self.image
		if image:
			r = image.get_rect()
			r.topright = rect.topright
			surface.blit(image, r)
	
	def thumbnail_size(self):
		return self.image.get_size()
	
	def draw_thumbnail(self, surface, rect):
		r = self.image.get_rect()
		r.center = rect.center
		surface.blit(self.image, r)
	
	def begin_frame(self, level, coords):
		pass
	
	def passable(self):
		return self.is_passable
	
	def player_has_entered(self):
		pass


class Wall(Item):

	is_climbable = False
	
	def _get_params(self, p):
		p.passable = self.is_passable
		p.climbable = self.is_climbable
	
	def _set_params(self, p):
		cls = self.__class__
		self.is_passable = getattr(p, 'passable', cls.is_passable)
		self.is_climbable = getattr(p, 'climbable', cls.is_climbable)

	def draw(self, surface, rect, front = False):
		if not front:
			image = self.image
			r = image.get_rect()
			r.bottomright = rect.bottomright
			surface.blit(image, r)
	
	def draw_behind(self, surface, rect):
		pass

	def climbable(self):
		return self.is_climbable
	
	def hit_by_player(self, player):
		pass
	
	def passed_by_player(self, player):
		pass


class Floor(Item):
	pass


class Back(Item):
	pass


class Gizmo(Item):

	height = None
	depth = None
	
	def _get_params(self, p):
		if self.height:
			p.height = self.height
		if self.depth:
			p.depth = self.depth

	def _set_params(self, p):
		self.height = getattr(p, 'height', None)
		self.depth = getattr(p, 'depth', None)

	def draw(self, surface, rect, front = False):
		if front:
			image = self.front_image
		else:
			image = self.image
		if image:
			r = image.get_rect()
			h = self.height or r.height
			d = self.depth or r.height - h
			x, y = rect.topright
			if self.depth and not self.height:
				y += self.depth
			r.midtop = (x - (cell_width + d) // 2, y + cell_height - h)
			surface.blit(image, r)
	
	def player_present(self, player):
		pass


class Detector(Gizmo):

	def _get_params(self, p):
		p.color = self.color
	
	def _set_params(self, p):
		self.color = p.color
	
	def player_present(self, player):
		if self.color <> player.cloak_tint():
			player.game.sound_alarm()


class Object(Item):

	value = 0
	description = ""
	
	def __init__(self, image, description = "", **kwds):
		Item.__init__(self, image = image, description = description, **kwds)
	
	def _get_params(self, p):
		p.description = self.description
	
	def _set_params(self, p):
		self.description = getattr(p, 'description', "")

	def draw(self, surface, rect, front = False):
		#print "Object.draw in", rect
		if not front:
			image = self.image
			r = image.get_rect()
			x, y = rect.bottomright
			r.midbottom = (x - cell_width // 2, y - cell_vdepth // 2)
			surface.blit(image, r)


class Battery(Object):

	kinds = ["red", "green", "blue"]

	def __init__(self, kind, image, description):
		Object.__init__(self, kind = kind, image = image,
			description = description)
	
	def _get_params(self, p):
		p.kind = self.kinds[self.kind]
	
	def _set_params(self, p):
		self.kind = self.kinds.index(p.kind)


class BeamWall(Wall):

	is_passable = True

	def __init__(self, color, image):
		Wall.__init__(self, color = color, image = image)
	
	def _get_params(self, p):
		p.color = self.color
	
	def _set_params(self, p):
		self.color = p.color
	
	def passed_by_player(self, player):
		if self.color <> player.cloak_tint():
			player.game.sound_alarm()


class Information(Object):

	is_static = False
	
	editor_attrs = [("Description", 'description'), ("Value", 'value', int)]
	
	def __init__(self, image, value = 0, description = ""):
		Object.__init__(self, image = image, value = value,
			description = description)

	def _get_params(self, p):
		p.value = self.value

	def _set_params(self, p):
		self.value = p.value
		self.description = p.description


register_class(Wall)
register_class(Floor)
register_class(Back)
register_class(Gizmo)
register_class(Object)

register_class(Battery)
register_class(BeamWall)
register_class(Information)
register_class(Detector)
