from Numeric import add, multiply
from pygame import Rect
from pygame.locals import *
from resource import get_image, get_sound
from level import cell_size, cell_depth, cell_hdepth, cell_vdepth, Battery
from doors import BackDoor, SideDoor
from stairs import Stairs
from animation import Animation

FACE_LEFT = 0
FACE_RIGHT = 1
FACE_BACK = 2

ACT_STAND = 'standing'
ACT_WALK = 'walking'
ACT_FALL = 'falling'
ACT_LEAVE = 'leaving'
ACT_CLIMB_WALL = 'climbing_wall'
ACT_CLIMB_STAIRS = 'climbing stairs'

CMD_NONE = 0
CMD_GO_LEFT = 1
CMD_GO_RIGHT = 2
CMD_GO_UP = 3
CMD_PICK_UP = 4
CMD_GO_DOWN = 5

key_to_cmd = {
	K_LEFT: CMD_GO_LEFT,
	K_RIGHT: CMD_GO_RIGHT,
	K_UP: CMD_GO_UP,
	K_DOWN: CMD_GO_DOWN,
	K_SPACE: CMD_PICK_UP,
}

player_height = 80
walk_speed = 10
climb_speed = 5
gravity = 6

cell_width, cell_height = cell_size
cell_xdepth, cell_ydepth = cell_depth

blue = (0, 0, 255)


class Player(object):
	#  position   (x, y)      of feet, from bottom left, x is right, y is up
	#  velocity   (vx, vy)    pixels per frame
	#  facing     FACE_*      facing direction
	#  command    CMD_*       commanded action
	#  anim       Animation   current animation sequence
	#  frame      int         current animation frame
	#  activity   ACT_*       current activity
	#  acting     boolean     true if performing activity
	#  inventory  [Object]    objects being carried
	#  charge     int[kind]   charge remaining in loaded batteries
	#  cloaks     bool[kind]  which cloaks are active
	#  cash       int         value of objects collected
	#  z          int         z-coordinate of position (0 == cell centre, + fwd)
	
	max_charge = 100
	
	standing = Animation("player_standing.png", 2, 1)
	walking = Animation("player_walking.png", 2, 8, border = 1)
	falling = Animation("player_falling.png", 2, 1)
	climbing_wall = Animation("player_climbing.png", 2, 8, border = 1)
	climbing_stairs = walking
	
	pickup_sound = get_sound("blip.wav")
	
	def __init__(self):
		self.position = (30, 0)
		self.velocity = (0, 0)
		self.facing = FACE_RIGHT
		self.activity = ACT_STAND
		self.acting = False
		self.command = CMD_NONE
		self.anim = self.standing
		self.frame = 0
		self.inventory = []
		self.charge = [0] * 3
		self.cloaks = [False] * 3
		self.cash = 0
		self.z = 0
	
	def draw(self, surface, level):
		px, py = self.position
		x = px + cell_xdepth / 2
		y = level.num_rows * cell_height - py + cell_ydepth / 2
		tint = self.cloak_tint()
		self.anim.draw_frame(surface, self.facing, self.frame,
			'midbottom', (x, y), tint)
	
	def cloak_tint(self):
		cloaks = self.cloaks
		if cloaks == [0, 0, 0]:
			return None
		else:
			return multiply(255, cloaks)

	def cell(self, level):
		return self.cell_coords(level)[0]

	def cell_coords(self, level):
		return level.actor_pos_to_cell_coords(self.position)
	
	def set_cell_coords(self, level, coords, offset):
		self.position = level.cell_coords_to_actor_pos(coords, offset)

	def key_down(self, event):
		#print "Player.key_down:", event
		key = event.key
		if key == K_r:
			self.toggle_cloak(0)
		elif key == K_g:
			self.toggle_cloak(1)
		elif key == K_b:
			self.toggle_cloak(2)
		cmd = key_to_cmd.get(key)
		if cmd:
			self.start_cmd(cmd)

	def key_up(self, event):
		#print "Player.key_up:", event
		cmd = key_to_cmd.get(event.key)
		if cmd:
			self.stop_cmd(cmd)
	
	def toggle_cloak(self, kind):
		cloaks = self.cloaks
		if not cloaks[kind] and self.load_charge(kind):
			cloaks[kind] = True
		else:
			cloaks[kind] = False
	
	def load_charge(self, kind):
		if self.charge[kind]:
			return True
		for item in self.inventory:
			if isinstance(item, Battery) and item.kind == kind:
				self.inventory.remove(item)
				self.charge[kind] = self.max_charge
				return True
		return False

	def start_cmd(self, cmd):
		self.command = cmd
	
	def stop_cmd(self, cmd):
		if self.command == cmd:
			self.command = CMD_NONE

	def begin_frame(self, level):
		self.diminish_charge()
		self.perform_action(level)
		self.determine_next_action(level)
	
	def diminish_charge(self):
		cloaks = self.cloaks
		charge = self.charge
		for i in range(3):
			if cloaks[i]:
				charge[i]-= 1
				if charge[i] == 0:
					if not self.load_charge(i):
						cloaks[i] = False
	
	def start_animation(self, anim):
		if self.anim is not anim:
			self.anim = anim
			self.frame = 0

	def perform_action(self, level):
		(row, col), (dx, dy) = self.cell_coords(level)
		gizmo = level.gizmo_at(row, col)
		if gizmo:
			gizmo.player_present(self)
		x, y = self.position
		vx, vy = self.velocity
		x2 = x
		y2 = y
		dx2 = dx
		dy2 = dy
		if self.acting:
			if self.activity == ACT_CLIMB_WALL and vy < 0:
				dir = -1
			else:
				dir = 1
			self.frame = (self.frame + dir) % self.anim.num_frames
			x2 += vx
			y2 += vy
			dx2 += vx
			dy2 += vy
			dx3 = dy3 = 0
			#print "Adding velocity %s to offset %s gives %s" % (
			#	(vx, vy), (dx, dy), (dx2, dy2))
		wall = floor = touching_floor = None
		passed_wall = passed_floor = hit_wall = hit_floor = None
		if dx2 < 0:
			wall = level.wall_at(row, col)
			dx3 = -dx2
		elif dx2 >= cell_width:
			wall = level.wall_at(row, col + 1)
			dx3 = cell_width - 1 - dx2
		if dy2 <= 0:
			floor = level.floor_at(row + 1, col)
			dy3 = -dy2
		elif dy2 + player_height >= cell_height:
			floor = level.floor_at(row, col)
			dy3 = cell_height - player_height - 1 - dy2
		if wall:
			if wall.passable():
				passed_wall = wall
			else:
				hit_wall = wall
				x2 += dx3
				#vx = 0
		if floor:
			if dy2 == 0:
				touching_floor = floor
			elif dy2 >= cell_height:
				passed_floor = floor
			elif floor.passable():
				if dy2 < 0:
					passed_floor = floor
			else:
				hit_floor = floor
				y2 += dy3
				#vy = 0
		if self.activity == ACT_FALL:
			vy -= gravity
			self.velocity = (vx, vy)
		#if passed_floor:
		#	print "Passed floor"
		#if hit_floor:
		#	print "Hit floor"
		self.position = (x2, y2)
		self.velocity = (vx, vy)
		self.touching_floor = touching_floor
		self.passed_wall = passed_wall
		self.passed_floor = passed_floor
		self.hit_wall = hit_wall
		self.hit_floor = hit_floor
		if passed_wall:
			passed_wall.passed_by_player(self)
		act = self.activity
		if act == ACT_LEAVE:
			self.do_leaving(level)

	def determine_next_action(self, level):
		act = self.activity
		cmd = self.command
		if act == ACT_STAND or act == ACT_WALK:
			if not self.touching_floor:
				self.fall()
			elif self.hit_wall:
				self.stand_still()
				self.hit_wall.hit_by_player(self)
			elif cmd == CMD_GO_LEFT:
				self.walk_left()
			elif cmd == CMD_GO_RIGHT:
				self.walk_right()
			elif cmd == CMD_GO_UP or cmd == CMD_GO_DOWN:
				if self.climbable_wall_in_front(level):
					self.start_climbing_wall(level)
				elif cmd == CMD_GO_UP and self.stairs_in_cell(level):
					self.start_climbing_up_stairs(level)
				elif cmd == CMD_GO_DOWN and self.stairs_in_cell_below(level):
					self.start_climbing_down_stairs(level)
				else:
					self.start_leaving(level)
			elif cmd == CMD_PICK_UP:
				self.do_pick_up(level)
			else:
				self.stand_still()
		elif act == ACT_CLIMB_WALL:
			if self.climbable_wall_in_front(level):
				if cmd == CMD_GO_UP:
					self.climb_up()
				elif cmd == CMD_GO_DOWN:
					self.climb_down()
				elif cmd == CMD_GO_LEFT:
					if self.facing == FACE_RIGHT:
						self.walk_left()
				elif cmd == CMD_GO_RIGHT:
					if self.facing == FACE_LEFT:
						self.walk_right()
				else:
					self.pause_climbing()
			else:
				self.stop_climbing(level)
		elif act == ACT_CLIMB_STAIRS:
			if self.hit_floor or self.passed_floor:
				#print "Reached floor while climbing stairs"
				self.stop_climbing_stairs(level)
		elif act == ACT_FALL:
			if self.touching_floor or self.hit_floor:
				self.stand_still()
				
	def walk_left(self):
		self.activity = ACT_WALK
		self.facing = FACE_LEFT
		self.velocity = (-walk_speed, 0)
		self.start_animation(self.walking)
		self.acting = True

	def walk_right(self):
		self.activity = ACT_WALK
		self.facing = FACE_RIGHT
		self.velocity = (walk_speed, 0)
		self.start_animation(self.walking)
		self.acting = True

	def stand_still(self):
		self.activity = ACT_STAND
		self.velocity = (0, 0)
		self.start_animation(self.standing)
		self.acting = False

	def fall(self):
		self.activity = ACT_FALL
		self.start_animation(self.falling)
		self.acting = True

	def start_leaving(self, level):
		row, col = self.cell(level)
		door = level.gizmo_at(row, col)
		print "Attempting to start leaving via", door
		if door and isinstance(door, BackDoor):
			down = self.command == CMD_GO_DOWN
			if door.get_link(down):
				print "Starting to leave"
				door.open()
				self.activity = ACT_LEAVE
				self.facing = FACE_BACK
				self.velocity = (0, 0)
				self.start_animation(self.standing)
				self.acting = False
				self.leaving_down = down
		self.command = None

	def do_leaving(self, level):
		print "Doing leaving with command", self.command
		row, col = self.cell(level)
		door = level.gizmo_at(row, col)
		if door and isinstance(door, BackDoor):
			print "Waiting to leave"
			if door.is_open():
				door.set_closed()
				link = door.get_link(down = self.leaving_down)
				print "Leaving via link", link
				self.game.switch_to_level(link)
				self.stand_still()
	
	def do_pick_up(self, level):
		row, col = self.cell(level)
		obj = level.take_object_at(row, col)
		if obj:
			self.inventory.insert(0, obj)
			self.cash += obj.value
			self.pickup_sound.play()
		self.cmd = None

	def start_climbing_wall(self, level):
		#print "Player.start_climbing"
		coords, (dx, dy) = self.cell_coords(level)
		self.activity = ACT_CLIMB_WALL
		if self.facing == FACE_LEFT:
			dx = 10
		else:
			dx = cell_width - 10
		self.set_cell_coords(level, coords, (dx, dy))
		self.start_animation(self.climbing_wall)
		self.acting = True
	
	def start_climbing_up_stairs(self,level):
		#print "Starting to climb up stairs"
		coords = self.cell(level)
		self.activity = ACT_CLIMB_STAIRS
		self.facing = FACE_LEFT
		self.set_cell_coords(level, coords, (cell_width - 10, 0))
		self.z = -cell_hdepth // 4
		self.velocity = (-5, 15)
		self.start_animation(self.climbing_stairs)
		self.acting = True
	
	def start_climbing_down_stairs(self,level):
		#print "Starting to climb down stairs"
		row, col = self.cell(level)
		self.activity = ACT_CLIMB_STAIRS
		self.facing = FACE_RIGHT
		self.velocity = (5, -15)
		self.set_cell_coords(level, (row + 1, col), (10, cell_height - 1))
		self.z = -cell_hdepth // 4
		self.start_animation(self.climbing_stairs)
		self.acting = True
	
	def stop_climbing_stairs(self, level):
		row, col = self.cell(level)
		#print "Stopping climbing stairs in cell", (row, col)
		vx, vy = self.velocity
		if vy < 0:
			self.facing = FACE_RIGHT
			if not self.hit_floor:
				#print "Going to row above"
				row -= 1
		else:
			self.facing = FACE_LEFT
		self.set_cell_coords(level, (row, col), (cell_width // 2, 0))
		self.z = 0
		self.stand_still()
	
	def climb_up(self):
		#print "Player.climb_up"
		self.velocity = (0, climb_speed)
		self.acting = True
	
	def climb_down(self):
		#print "Player.climb_down"
		self.velocity = (0, -climb_speed)
		self.acting = True
	
	def pause_climbing(self):
		#print "Player.pause_climbing"
		self.velocity = (0, 0)
		self.acting = False
	
	def stop_climbing(self, level):
		(row, col), (dx, dy) = self.cell_coords(level)
		if self.facing == FACE_LEFT:
			col -= 1
			dx = cell_width - 10
		else:
			col += 1
			dx = 10
		self.set_cell_coords(level, (row, col), (dx, dy))
		self.stand_still()
		self.command = CMD_NONE

	def wall_in_front(self, level):
		row, col = self.cell(level)
		if self.facing == FACE_RIGHT:
			col += 1
		wall = level.wall_at(row, col)
		#print "Player.wall_in_front =", wall
		return wall

	def climbable_wall_in_front(self, level):
		wall = self.wall_in_front(level)
		return wall and wall.climbable()

	def stairs_in_cell(self, level):
		row, col = self.cell(level)
		item = level.gizmo_at(row, col)
		return item and isinstance(item, Stairs)

	def stairs_in_cell_below(self, level):
		row, col = self.cell(level)
		item = level.gizmo_at(row + 1, col)
		return isinstance(item, Stairs)
